from .base import ModuleTestBase


class TestInternetDB(ModuleTestBase):
    config_overrides = {"dns_resolution": True}

    async def setup_before_prep(self, module_test):
        module_test.scan.helpers.mock_dns(
            {
                ("blacklanternsecurity.com", "A"): "1.2.3.4",
                ("autodiscover.blacklanternsecurity.com", "A"): "2.3.4.5",
                ("mail.blacklanternsecurity.com", "A"): "3.4.5.6",
            }
        )

        module_test.httpx_mock.add_response(
            url="https://internetdb.shodan.io/1.2.3.4",
            json={
                "cpes": [
                    "cpe:/a:microsoft:internet_information_services",
                    "cpe:/a:microsoft:outlook_web_access:15.0.1367",
                ],
                "hostnames": [
                    "autodiscover.blacklanternsecurity.com",
                    "mail.blacklanternsecurity.com",
                ],
                "ip": "1.2.3.4",
                "ports": [
                    25,
                    80,
                    443,
                ],
                "tags": ["starttls", "self-signed", "eol-os"],
                "vulns": ["CVE-2021-26857", "CVE-2021-26855"],
            },
        )

    def check(self, module_test, events):
        assert 5 == len([e for e in events if str(e.module) == "internetdb"])
        assert 1 == len(
            [e for e in events if e.type == "DNS_NAME" and e.data == "autodiscover.blacklanternsecurity.com"]
        )
        assert 1 == len([e for e in events if e.type == "DNS_NAME" and e.data == "mail.blacklanternsecurity.com"])
        assert 0 == len([e for e in events if e.type == "OPEN_TCP_PORT"])
        assert 1 == len([e for e in events if e.type == "FINDING" and str(e.module) == "internetdb"])
        assert 1 == len([e for e in events if e.type == "FINDING" and "CVE-2021-26857" in e.data["description"]])
        assert 2 == len([e for e in events if e.type == "TECHNOLOGY" and str(e.module) == "internetdb"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "TECHNOLOGY" and e.data["technology"] == "cpe:/a:microsoft:outlook_web_access:15.0.1367"
            ]
        )



from .base import ModuleTestBase


class TestPostman(ModuleTestBase):
    config_overrides = {
        "omit_event_types": [],
        "scope_report_distance": 1,
    }

    modules_overrides = ["postman", "httpx", "excavate"]

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/ws/proxy",
            json={
                "data": [
                    {
                        "score": 499.22498,
                        "normalizedScore": 8.43312276976538,
                        "document": {
                            "isPublisherVerified": False,
                            "publisherType": "user",
                            "curatedInList": [],
                            "publisherId": "28329861",
                            "publisherHandle": "",
                            "publisherLogo": "",
                            "isPublic": True,
                            "customHostName": "",
                            "id": "28329861-28329861-f6ef-4f23-9f3a-8431f3567ac1",
                            "workspaces": [
                                {
                                    "visibilityStatus": "public",
                                    "name": "BlackLanternSecuritySpilledSecrets",
                                    "id": "afa061be-9cb0-4520-9d4d-fe63361daf0f",
                                    "slug": "blacklanternsecurityspilledsecrets",
                                }
                            ],
                            "collectionForkLabel": "",
                            "method": "POST",
                            "entityType": "request",
                            "url": "www.example.com/index",
                            "isBlacklisted": False,
                            "warehouse__updated_at_collection": "2023-12-11 02:00:00",
                            "isPrivateNetworkEntity": False,
                            "warehouse__updated_at_request": "2023-12-11 02:00:00",
                            "publisherName": "NA",
                            "name": "A test post request",
                            "privateNetworkMeta": "",
                            "privateNetworkFolders": [],
                            "documentType": "request",
                            "collection": {
                                "id": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                "name": "Secret Collection",
                                "tags": [],
                                "forkCount": 0,
                                "watcherCount": 0,
                                "views": 31,
                                "apiId": "",
                                "apiName": "",
                            },
                        },
                    },
                    {
                        "score": 498.22398,
                        "normalizedScore": 8.43312266976538,
                        "document": {
                            "isPublisherVerified": False,
                            "publisherType": "user",
                            "curatedInList": [],
                            "publisherId": "28329861",
                            "publisherHandle": "",
                            "publisherLogo": "",
                            "isPublic": True,
                            "customHostName": "",
                            "id": "b7fa2137-b7fa2137-23bf-45d1-b176-35359af30ded",
                            "workspaces": [
                                {
                                    "visibilityStatus": "public",
                                    "name": "SpilledSecrets",
                                    "id": "92d0451b-119d-4ef0-b74c-22c400e5ce05",
                                    "slug": "spilledsecrets",
                                }
                            ],
                            "collectionForkLabel": "",
                            "method": "POST",
                            "entityType": "request",
                            "url": "www.example.com/index",
                            "isBlacklisted": False,
                            "warehouse__updated_at_collection": "2023-12-11 02:00:00",
                            "isPrivateNetworkEntity": False,
                            "warehouse__updated_at_request": "2023-12-11 02:00:00",
                            "publisherName": "NA",
                            "name": "A test post request",
                            "privateNetworkMeta": "",
                            "privateNetworkFolders": [],
                            "documentType": "request",
                            "collection": {
                                "id": "007e8d67-007e8d67-932b-46ff-b95c-a2aa216edaf3",
                                "name": "Secret Collection",
                                "tags": [],
                                "forkCount": 0,
                                "watcherCount": 0,
                                "views": 31,
                                "apiId": "",
                                "apiName": "",
                            },
                        },
                    },
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/workspace/afa061be-9cb0-4520-9d4d-fe63361daf0f",
            json={
                "model_id": "afa061be-9cb0-4520-9d4d-fe63361daf0f",
                "meta": {"model": "workspace", "action": "find"},
                "data": {
                    "id": "afa061be-9cb0-4520-9d4d-fe63361daf0f",
                    "name": "SpilledSecrets",
                    "description": "A Mock workspace environment filled with fake secrets to act as a testing ground for secret scanners",
                    "summary": "A Public workspace with mock secrets",
                    "createdBy": "28329861",
                    "updatedBy": "28329861",
                    "team": "28329861",
                    "createdAt": "2023-12-13T19:12:21.000Z",
                    "updatedAt": "2023-12-13T19:15:07.000Z",
                    "visibilityStatus": "public",
                    "profileInfo": {
                        "slug": "spilledsecrets",
                        "profileType": "team",
                        "profileId": "28329861",
                        "publicHandle": "https://www.postman.com/lunar-pizza-28329861",
                        "publicImageURL": "https://res.cloudinary.com/postman/image/upload/t_team_logo/v1/team/default-L4",
                        "publicName": "lunar-pizza-28329861",
                        "isVerified": False,
                    },
                    "user": "28329861",
                    "type": "team",
                    "dependencies": {
                        "collections": ["28129865-d9f8833b-3dd2-4b07-9634-1831206d5205"],
                        "environments": ["28129865-fa7edca0-2df6-4187-9805-11845912f567"],
                        "globals": ["28735eff-5ecd-43e6-8473-387f160f220f"],
                    },
                    "members": {"users": {"28129865": {"id": "28129865"}}},
                },
            },
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/list/collection?workspace=afa061be-9cb0-4520-9d4d-fe63361daf0f",
            json={
                "data": [
                    {
                        "id": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                        "name": "Secret Collection",
                        "folders_order": ["d6dd1092-94d4-462c-be48-4e3cad3b8612"],
                        "order": ["67c9db4c-d0ed-461c-86d2-9a8c5a5de896"],
                        "attributes": {
                            "permissions": {
                                "userCanUpdate": False,
                                "userCanDelete": False,
                                "userCanShare": False,
                                "userCanCreateMock": False,
                                "userCanCreateMonitor": False,
                                "anybodyCanView": True,
                                "teamCanView": True,
                            },
                            "fork": None,
                            "parent": {"type": "workspace", "id": "afa061be-9cb0-4520-9d4d-fe63361daf0f"},
                            "flags": {"isArchived": False, "isFavorite": False},
                        },
                        "folders": [
                            {
                                "id": "28129865-d6dd1092-94d4-462c-be48-4e3cad3b8612",
                                "name": "Nested folder",
                                "folder": None,
                                "collection": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                "folders_order": ["73a999c1-4246-4805-91bd-0232cc75958a"],
                                "order": ["3aa78b71-2c4f-4299-94df-287ed1036409"],
                                "folders": [
                                    {
                                        "id": "28129865-73a999c1-4246-4805-91bd-0232cc75958a",
                                        "name": "Another Nested Folder",
                                        "folder": "28129865-d6dd1092-94d4-462c-be48-4e3cad3b8612",
                                        "collection": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                        "folders_order": [],
                                        "order": ["987c8ac8-bfa9-4bab-ade9-88ccf0597862"],
                                        "folders": [],
                                        "requests": [
                                            {
                                                "id": "28129865-987c8ac8-bfa9-4bab-ade9-88ccf0597862",
                                                "name": "Delete User",
                                                "method": "DELETE",
                                                "collection": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                                "folder": "28129865-73a999c1-4246-4805-91bd-0232cc75958a",
                                                "responses_order": [],
                                                "responses": [],
                                            }
                                        ],
                                    }
                                ],
                                "requests": [
                                    {
                                        "id": "28129865-3aa78b71-2c4f-4299-94df-287ed1036409",
                                        "name": "Login",
                                        "method": "POST",
                                        "collection": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                        "folder": "28129865-d6dd1092-94d4-462c-be48-4e3cad3b8612",
                                        "responses_order": [],
                                        "responses": [],
                                    }
                                ],
                            }
                        ],
                        "requests": [
                            {
                                "id": "28129865-67c9db4c-d0ed-461c-86d2-9a8c5a5de896",
                                "name": "Example Basic HTTP request",
                                "method": "GET",
                                "collection": "28129865-d9f8833b-3dd2-4b07-9634-1831206d5205",
                                "folder": None,
                                "responses_order": [],
                                "responses": [],
                            }
                        ],
                    }
                ]
            },
        )

        old_emit_event = module_test.module.emit_event

        async def new_emit_event(event_data, event_type, **kwargs):
            if event_data.startswith("https://www.postman.com"):
                event_data = event_data.replace("https://www.postman.com", "http://127.0.0.1:8888")
            await old_emit_event(event_data, event_type, **kwargs)

        module_test.monkeypatch.setattr(module_test.module, "emit_event", new_emit_event)
        module_test.scan.helpers.dns.mock_dns({("asdf.blacklanternsecurity.com", "A"): "127.0.0.1"})

        request_args = dict(uri="/_api/request/28129865-987c8ac8-bfa9-4bab-ade9-88ccf0597862")
        respond_args = dict(response_data="https://asdf.blacklanternsecurity.com")
        module_test.set_expect_requests(request_args, respond_args)

    def check(self, module_test, events):
        assert any(
            e.data == "http://127.0.0.1:8888/_api/workspace/afa061be-9cb0-4520-9d4d-fe63361daf0f" for e in events
        ), "Failed to detect workspace"
        assert any(
            e.data != "http://127.0.0.1:8888/_api/workspace/92d0451b-119d-4ef0-b74c-22c400e5ce05" for e in events
        ), "Workspace should not be detected"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/workspace/afa061be-9cb0-4520-9d4d-fe63361daf0f/globals"
            for e in events
        ), "Failed to detect workspace globals"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/environment/28129865-fa7edca0-2df6-4187-9805-11845912f567"
            for e in events
        ), "Failed to detect workspace environment"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/collection/28129865-d9f8833b-3dd2-4b07-9634-1831206d5205"
            for e in events
        ), "Failed to detect collection"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/request/28129865-987c8ac8-bfa9-4bab-ade9-88ccf0597862"
            for e in events
        ), "Failed to detect collection request #1"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/request/28129865-3aa78b71-2c4f-4299-94df-287ed1036409"
            for e in events
        ), "Failed to detect collection request #2"
        assert any(
            e.data == "http://127.0.0.1:8888/_api/request/28129865-67c9db4c-d0ed-461c-86d2-9a8c5a5de896"
            for e in events
        ), "Failed to detect collection request #3"
        assert any(
            e.type == "HTTP_RESPONSE"
            and e.data["url"] == "http://127.0.0.1:8888/_api/request/28129865-987c8ac8-bfa9-4bab-ade9-88ccf0597862"
            for e in events
        ), "Failed to emit HTTP_RESPONSE"
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_dns(bbot_scanner, bbot_config):
    scan = bbot_scanner("1.1.1.1", config=bbot_config)
    helpers = scan.helpers

    # lowest level functions
    a_responses = await helpers._resolve_hostname("one.one.one.one")
    aaaa_responses = await helpers._resolve_hostname("one.one.one.one", rdtype="AAAA")
    ip_responses = await helpers._resolve_ip("1.1.1.1")
    assert a_responses[0].response.answer[0][0].address in ("1.1.1.1", "1.0.0.1")
    assert aaaa_responses[0].response.answer[0][0].address in ("2606:4700:4700::1111", "2606:4700:4700::1001")
    assert ip_responses[0].response.answer[0][0].target.to_text() in ("one.one.one.one.",)

    # mid level functions
    _responses, errors = await helpers.resolve_raw("one.one.one.one")
    responses = []
    for rdtype, response in _responses:
        for answers in response:
            responses += list(helpers.extract_targets(answers))
    assert ("A", "1.1.1.1") in responses
    _responses, errors = await helpers.resolve_raw("one.one.one.one", rdtype="AAAA")
    responses = []
    for rdtype, response in _responses:
        for answers in response:
            responses += list(helpers.extract_targets(answers))
    assert ("AAAA", "2606:4700:4700::1111") in responses
    _responses, errors = await helpers.resolve_raw("1.1.1.1")
    responses = []
    for rdtype, response in _responses:
        for answers in response:
            responses += list(helpers.extract_targets(answers))
    assert ("PTR", "one.one.one.one") in responses

    # high level functions
    assert "1.1.1.1" in await helpers.resolve("one.one.one.one")
    assert "2606:4700:4700::1111" in await helpers.resolve("one.one.one.one", type="AAAA")
    assert "one.one.one.one" in await helpers.resolve("1.1.1.1")
    for rdtype in ("NS", "SOA", "MX", "TXT"):
        assert len(await helpers.resolve("google.com", type=rdtype)) > 0

    # batch resolution
    batch_results = [r async for r in helpers.resolve_batch(["1.1.1.1", "one.one.one.one"])]
    assert len(batch_results) == 2
    batch_results = dict(batch_results)
    assert any([x in batch_results["one.one.one.one"] for x in ("1.1.1.1", "1.0.0.1")])
    assert "one.one.one.one" in batch_results["1.1.1.1"]

    # "any" type
    resolved = await helpers.resolve("google.com", type="any")
    assert any([helpers.is_subdomain(h) for h in resolved])

    # dns cache
    helpers.dns._dns_cache.clear()
    assert hash(f"1.1.1.1:PTR") not in helpers.dns._dns_cache
    assert hash(f"one.one.one.one:A") not in helpers.dns._dns_cache
    assert hash(f"one.one.one.one:AAAA") not in helpers.dns._dns_cache
    await helpers.resolve("1.1.1.1", use_cache=False)
    await helpers.resolve("one.one.one.one", use_cache=False)
    assert hash(f"1.1.1.1:PTR") not in helpers.dns._dns_cache
    assert hash(f"one.one.one.one:A") not in helpers.dns._dns_cache
    assert hash(f"one.one.one.one:AAAA") not in helpers.dns._dns_cache

    await helpers.resolve("1.1.1.1")
    assert hash(f"1.1.1.1:PTR") in helpers.dns._dns_cache
    await helpers.resolve("one.one.one.one")
    assert hash(f"one.one.one.one:A") in helpers.dns._dns_cache
    assert hash(f"one.one.one.one:AAAA") in helpers.dns._dns_cache

    # Ensure events with hosts have resolved_hosts attribute populated
    resolved_hosts_event1 = scan.make_event("one.one.one.one", "DNS_NAME", dummy=True)
    resolved_hosts_event2 = scan.make_event("http://one.one.one.one/", "URL_UNVERIFIED", dummy=True)
    event_tags1, event_whitelisted1, event_blacklisted1, children1 = await scan.helpers.resolve_event(
        resolved_hosts_event1
    )
    event_tags2, event_whitelisted2, event_blacklisted2, children2 = await scan.helpers.resolve_event(
        resolved_hosts_event2
    )
    assert "1.1.1.1" in [str(x) for x in children1["A"]]
    assert "1.1.1.1" in [str(x) for x in children2["A"]]
    assert set(children1.keys()) == set(children2.keys())

    dns_config = OmegaConf.create({"dns_resolution": True})
    dns_config = OmegaConf.merge(bbot_config, dns_config)
    scan2 = bbot_scanner("evilcorp.com", config=dns_config)
    scan2.helpers.dns.mock_dns(
        {("evilcorp.com", "TXT"): '"v=spf1 include:cloudprovider.com ~all"', ("cloudprovider.com", "A"): "1.2.3.4"}
    )
    events = [e async for e in scan2.async_start()]
    assert 1 == len(
        [e for e in events if e.type == "DNS_NAME" and e.data == "cloudprovider.com" and "affiliate" in e.tags]
    )


@pytest.mark.asyncio
async def test_wildcards(bbot_scanner, bbot_config):
    scan = bbot_scanner("1.1.1.1", config=bbot_config)
    helpers = scan.helpers

    # wildcards
    wildcard_domains = await helpers.is_wildcard_domain("asdf.github.io")
    assert hash("github.io") in helpers.dns._wildcard_cache
    assert hash("asdf.github.io") in helpers.dns._wildcard_cache
    assert "github.io" in wildcard_domains
    assert "A" in wildcard_domains["github.io"]
    assert "SRV" not in wildcard_domains["github.io"]
    assert wildcard_domains["github.io"]["A"] and all(helpers.is_ip(r) for r in wildcard_domains["github.io"]["A"])
    helpers.dns._wildcard_cache.clear()

    wildcard_rdtypes = await helpers.is_wildcard("blacklanternsecurity.github.io")
    assert "A" in wildcard_rdtypes
    assert "SRV" not in wildcard_rdtypes
    assert wildcard_rdtypes["A"] == (True, "github.io")
    assert hash("github.io") in helpers.dns._wildcard_cache
    assert len(helpers.dns._wildcard_cache[hash("github.io")]) > 0
    helpers.dns._wildcard_cache.clear()

    wildcard_rdtypes = await helpers.is_wildcard("asdf.asdf.asdf.github.io")
    assert "A" in wildcard_rdtypes
    assert "SRV" not in wildcard_rdtypes
    assert wildcard_rdtypes["A"] == (True, "github.io")
    assert hash("github.io") in helpers.dns._wildcard_cache
    assert not hash("asdf.github.io") in helpers.dns._wildcard_cache
    assert not hash("asdf.asdf.github.io") in helpers.dns._wildcard_cache
    assert not hash("asdf.asdf.asdf.github.io") in helpers.dns._wildcard_cache
    assert len(helpers.dns._wildcard_cache[hash("github.io")]) > 0
    wildcard_event1 = scan.make_event("wat.asdf.fdsa.github.io", "DNS_NAME", dummy=True)
    wildcard_event2 = scan.make_event("wats.asd.fdsa.github.io", "DNS_NAME", dummy=True)
    wildcard_event3 = scan.make_event("github.io", "DNS_NAME", dummy=True)

    # event resolution
    event_tags1, event_whitelisted1, event_blacklisted1, children1 = await scan.helpers.resolve_event(wildcard_event1)
    event_tags2, event_whitelisted2, event_blacklisted2, children2 = await scan.helpers.resolve_event(wildcard_event2)
    event_tags3, event_whitelisted3, event_blacklisted3, children3 = await scan.helpers.resolve_event(wildcard_event3)
    await helpers.handle_wildcard_event(wildcard_event1, children1)
    await helpers.handle_wildcard_event(wildcard_event2, children2)
    await helpers.handle_wildcard_event(wildcard_event3, children3)
    assert "wildcard" in wildcard_event1.tags
    assert "a-wildcard" in wildcard_event1.tags
    assert "srv-wildcard" not in wildcard_event1.tags
    assert "wildcard" in wildcard_event2.tags
    assert "a-wildcard" in wildcard_event2.tags
    assert "srv-wildcard" not in wildcard_event2.tags
    assert wildcard_event1.data == "_wildcard.github.io"
    assert wildcard_event2.data == "_wildcard.github.io"
    assert wildcard_event1.tags == wildcard_event2.tags
    assert "wildcard-domain" in wildcard_event3.tags
    assert "a-wildcard-domain" in wildcard_event3.tags
    assert "srv-wildcard-domain" not in wildcard_event3.tags



from .base import ModuleTestBase


class TestAggregate(ModuleTestBase):
    config_overrides = {"dns_resolution": True, "scope_report_distance": 1}

    async def setup_before_prep(self, module_test):
        module_test.scan.helpers.dns.mock_dns({("blacklanternsecurity.com", "A"): "1.2.3.4"})

    def check(self, module_test, events):
        filename = next(module_test.scan.home.glob("scan-stats-table*.txt"))
        with open(filename) as f:
            assert "| A  " in f.read()



from .base import ModuleTestBase


class TestGithub_Codesearch(ModuleTestBase):
    config_overrides = {
        "modules": {"github_codesearch": {"api_key": "asdf", "limit": 1}},
        "omit_event_types": [],
        "scope_report_distance": 1,
    }
    modules_overrides = ["github_codesearch", "httpx", "secretsdb"]

    github_file_endpoint = (
        "/projectdiscovery/nuclei/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go"
    )
    github_file_url = f"http://127.0.0.1:8888{github_file_endpoint}"

    async def setup_before_prep(self, module_test):
        expect_args = {"method": "GET", "uri": self.github_file_endpoint}
        respond_args = {"response_data": "-----BEGIN PGP PRIVATE KEY BLOCK-----"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        module_test.httpx_mock.add_response(url="https://api.github.com/zen")
        module_test.httpx_mock.add_response(
            url="https://api.github.com/search/code?per_page=100&type=Code&q=blacklanternsecurity.com&page=1",
            json={
                "total_count": 214,
                "incomplete_results": False,
                "items": [
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go2",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go3",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                ],
            },
        )

    async def setup_after_prep(self, module_test):
        module_test.module.github_raw_url = "http://127.0.0.1:8888/"

    def check(self, module_test, events):
        assert 1 == len([e for e in events if e.type == "URL_UNVERIFIED"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL_UNVERIFIED" and e.data == self.github_file_url and e.scope_distance == 1
            ]
        ), "Failed to emit URL_UNVERIFIED"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and e.data["url"] == "https://github.com/projectdiscovery/nuclei"
                and e.scope_distance == 1
            ]
        ), "Failed to emit CODE_REPOSITORY"
        assert 1 == len(
            [e for e in events if e.type == "URL" and e.data == self.github_file_url and e.scope_distance == 1]
        ), "Failed to visit URL"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HTTP_RESPONSE" and e.data["url"] == self.github_file_url and e.scope_distance == 1
            ]
        ), "Failed to visit URL"
        assert [
            e for e in events if e.type == "FINDING" and str(e.module) == "secretsdb"
        ], "Failed to find secret in repo file"



import sqlite3
from pathlib import Path
from contextlib import suppress
from shutil import copyfile, copymode

from bbot.modules.base import BaseModule


class gowitness(BaseModule):
    watched_events = ["URL", "SOCIAL"]
    produced_events = ["WEBSCREENSHOT", "URL", "URL_UNVERIFIED", "TECHNOLOGY"]
    flags = ["active", "safe", "web-screenshots"]
    meta = {"description": "Take screenshots of webpages"}
    options = {
        "version": "2.4.2",
        "threads": 4,
        "timeout": 10,
        "resolution_x": 1440,
        "resolution_y": 900,
        "output_path": "",
    }
    options_desc = {
        "version": "gowitness version",
        "threads": "threads used to run",
        "timeout": "preflight check timeout",
        "resolution_x": "screenshot resolution x",
        "resolution_y": "screenshot resolution y",
        "output_path": "where to save screenshots",
    }
    deps_ansible = [
        {
            "name": "Install Chromium (Non-Debian)",
            "package": {"name": "chromium", "state": "present"},
            "become": True,
            "when": "ansible_facts['os_family'] != 'Debian'",
            "ignore_errors": True,
        },
        {
            "name": "Install Chromium dependencies (Debian)",
            "package": {
                "name": "libasound2,libatk-bridge2.0-0,libatk1.0-0,libcairo2,libcups2,libdrm2,libgbm1,libnss3,libpango-1.0-0,libxcomposite1,libxdamage1,libxfixes3,libxkbcommon0,libxrandr2",
                "state": "present",
            },
            "become": True,
            "when": "ansible_facts['os_family'] == 'Debian'",
            "ignore_errors": True,
        },
        {
            "name": "Get latest Chromium version (Debian)",
            "uri": {
                "url": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Linux_x64%2FLAST_CHANGE?alt=media",
                "return_content": True,
            },
            "register": "chromium_version",
            "when": "ansible_facts['os_family'] == 'Debian'",
            "ignore_errors": True,
        },
        {
            "name": "Download Chromium (Debian)",
            "unarchive": {
                "src": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Linux_x64%2F{{ chromium_version.content }}%2Fchrome-linux.zip?alt=media",
                "remote_src": True,
                "dest": "#{BBOT_TOOLS}",
                "creates": "#{BBOT_TOOLS}/chrome-linux",
            },
            "when": "ansible_facts['os_family'] == 'Debian'",
            "ignore_errors": True,
        },
        {
            "name": "Download gowitness",
            "get_url": {
                "url": "https://github.com/sensepost/gowitness/releases/download/#{BBOT_MODULES_GOWITNESS_VERSION}/gowitness-#{BBOT_MODULES_GOWITNESS_VERSION}-#{BBOT_OS_PLATFORM}-#{BBOT_CPU_ARCH}",
                "dest": "#{BBOT_TOOLS}/gowitness",
                "mode": "755",
            },
        },
    ]
    _batch_size = 100
    # gowitness accepts SOCIAL events up to distance 2, otherwise it is in-scope-only
    scope_distance_modifier = 2

    async def setup(self):
        self.timeout = self.config.get("timeout", 10)
        self.threads = self.config.get("threads", 4)
        self.proxy = self.scan.config.get("http_proxy", "")
        self.resolution_x = self.config.get("resolution_x")
        self.resolution_y = self.config.get("resolution_y")
        output_path = self.config.get("output_path")
        if output_path:
            self.base_path = Path(output_path) / "gowitness"
        else:
            self.base_path = self.scan.home / "gowitness"
        self.chrome_path = None
        custom_chrome_path = self.helpers.tools_dir / "chrome-linux" / "chrome"
        if custom_chrome_path.is_file():
            self.chrome_path = custom_chrome_path
        self.db_path = self.base_path / "gowitness.sqlite3"
        self.screenshot_path = self.base_path / "screenshots"
        self.command = self.construct_command()
        self.prepped = False
        self.screenshots_taken = dict()
        self.connections_logged = set()
        self.technologies_found = set()
        return True

    def prep(self):
        if not self.prepped:
            self.helpers.mkdir(self.screenshot_path)
            self.db_path.touch()
            with suppress(Exception):
                copyfile(self.helpers.tools_dir / "gowitness", self.base_path / "gowitness")
                copymode(self.helpers.tools_dir / "gowitness", self.base_path / "gowitness")
            self.prepped = True

    async def filter_event(self, event):
        # Ignore URLs that are redirects
        if any(t.startswith("status-30") for t in event.tags):
            return False, "URL is a redirect"
        # ignore events from self
        if event.type == "URL" and event.module == self:
            return False, "event is from self"
        # Accept out-of-scope SOCIAL pages, but not URLs
        if event.scope_distance > 0:
            if event.type != "SOCIAL":
                return False, "event is not in-scope"
        return True

    async def handle_batch(self, *events):
        self.prep()
        event_dict = {}
        for e in events:
            key = e.data
            if e.type == "SOCIAL":
                key = e.data["url"]
            event_dict[key] = e
        stdin = "\n".join(list(event_dict))

        async for line in self.helpers.run_live(self.command, input=stdin):
            self.debug(line)

        # emit web screenshots
        for filename, screenshot in self.new_screenshots.items():
            url = screenshot["url"]
            final_url = screenshot["final_url"]
            filename = screenshot["filename"]
            webscreenshot_data = {"filename": filename, "url": final_url}
            source_event = event_dict[url]
            await self.emit_event(webscreenshot_data, "WEBSCREENSHOT", source=source_event)

        # emit URLs
        for url, row in self.new_network_logs.items():
            ip = row["ip"]
            status_code = row["status_code"]
            tags = [f"status-{status_code}", f"ip-{ip}"]

            _id = row["url_id"]
            source_url = self.screenshots_taken[_id]
            source_event = event_dict[source_url]
            if self.helpers.is_spider_danger(source_event, url):
                tags.append("spider-danger")
            if url and url.startswith("http"):
                await self.emit_event(url, "URL_UNVERIFIED", source=source_event, tags=tags)

        # emit technologies
        for _, row in self.new_technologies.items():
            source_id = row["url_id"]
            source_url = self.screenshots_taken[source_id]
            source_event = event_dict[source_url]
            technology = row["value"]
            tech_data = {"technology": technology, "url": source_url, "host": str(source_event.host)}
            await self.emit_event(tech_data, "TECHNOLOGY", source=source_event)

    def construct_command(self):
        # base executable
        command = ["gowitness"]
        # chrome path
        if self.chrome_path is not None:
            command += ["--chrome-path", str(self.chrome_path)]
        # db path
        command += ["--db-path", str(self.db_path)]
        # screenshot path
        command += ["--screenshot-path", str(self.screenshot_path)]
        # user agent
        command += ["--user-agent", f"{self.scan.useragent}"]
        # proxy
        if self.proxy:
            command += ["--proxy", str(self.proxy)]
        # resolution
        command += ["--resolution-x", str(self.resolution_x)]
        command += ["--resolution-y", str(self.resolution_y)]
        # input
        command += ["file", "-f", "-"]
        # threads
        command += ["--threads", str(self.threads)]
        return command

    @property
    def new_screenshots(self):
        screenshots = {}
        if self.db_path.is_file():
            with sqlite3.connect(str(self.db_path)) as con:
                con.row_factory = sqlite3.Row
                cur = con.cursor()
                res = self.cur_execute(cur, "SELECT * FROM urls")
                for row in res:
                    row = dict(row)
                    _id = row["id"]
                    if _id not in self.screenshots_taken:
                        self.screenshots_taken[_id] = row["url"]
                        screenshots[_id] = row
        return screenshots

    @property
    def new_network_logs(self):
        network_logs = dict()
        if self.db_path.is_file():
            with sqlite3.connect(str(self.db_path)) as con:
                con.row_factory = sqlite3.Row
                cur = con.cursor()
                res = self.cur_execute(cur, "SELECT * FROM network_logs")
                for row in res:
                    row = dict(row)
                    url = row["final_url"]
                    if url not in self.connections_logged:
                        self.connections_logged.add(url)
                        network_logs[url] = row
        return network_logs

    @property
    def new_technologies(self):
        technologies = dict()
        if self.db_path.is_file():
            with sqlite3.connect(str(self.db_path)) as con:
                con.row_factory = sqlite3.Row
                cur = con.cursor()
                res = self.cur_execute(cur, "SELECT * FROM technologies")
                for row in res:
                    _id = row["id"]
                    if _id not in self.technologies_found:
                        self.technologies_found.add(_id)
                        row = dict(row)
                        technologies[_id] = row
        return technologies

    def cur_execute(self, cur, query):
        try:
            return cur.execute(query)
        except sqlite3.OperationalError as e:
            self.warning(f"Error executing query: {query}: {e}")
            return []

    async def report(self):
        if self.screenshots_taken:
            self.success(f"{len(self.screenshots_taken):,} web screenshots captured. To view:")
            self.success(f"    - Start gowitness")
            self.success(f"        - cd {self.base_path} && ./gowitness server")
            self.success(f"    - Browse to http://localhost:7171")
        else:
            self.info(f"No web screenshots captured")



import logging
import traceback

log = logging.getLogger("bbot.scanner.dispatcher")


class Dispatcher:
    """
    Enables custom hooks/callbacks on certain scan events
    """

    def set_scan(self, scan):
        self.scan = scan

    async def on_start(self, scan):
        return

    async def on_finish(self, scan):
        return

    async def on_status(self, status, scan_id):
        """
        Execute an event when the scan's status is updated
        """
        self.scan.debug(f"Setting scan status to {status}")

    async def catch(self, callback, *args, **kwargs):
        try:
            return await callback(*args, **kwargs)
        except Exception as e:
            log.error(f"Error in {callback.__qualname__}(): {e}")
            log.trace(traceback.format_exc())



import os
import sys
import logging
from copy import copy
import logging.handlers
from pathlib import Path

from ..configurator import config
from ..helpers.misc import mkdir, error_and_exit
from ..helpers.logger import colorize, loglevel_mapping


_log_level_override = None

bbot_loggers = None
bbot_log_handlers = None

debug_format = logging.Formatter("%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)s %(message)s")


class ColoredFormatter(logging.Formatter):
    """
    Pretty colors for terminal
    """

    formatter = logging.Formatter("%(levelname)s %(message)s")
    module_formatter = logging.Formatter("%(levelname)s %(name)s: %(message)s")

    def format(self, record):
        colored_record = copy(record)
        levelname = colored_record.levelname
        levelshort = loglevel_mapping.get(levelname, "INFO")
        colored_record.levelname = colorize(f"[{levelshort}]", level=levelname)
        if levelname == "CRITICAL" or levelname.startswith("HUGE"):
            colored_record.msg = colorize(colored_record.msg, level=levelname)
        # remove name
        if colored_record.name.startswith("bbot.modules."):
            colored_record.name = colored_record.name.split("bbot.modules.")[-1]
            return self.module_formatter.format(colored_record)
        return self.formatter.format(colored_record)


def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel('TRACE')
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError(f"{levelName} already defined in logging module")
    if hasattr(logging, methodName):
        raise AttributeError(f"{methodName} already defined in logging module")
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError(f"{methodName} already defined in logger class")

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


# custom logging levels
addLoggingLevel("STDOUT", 100)
addLoggingLevel("TRACE", 49)
addLoggingLevel("HUGEWARNING", 31)
addLoggingLevel("HUGESUCCESS", 26)
addLoggingLevel("SUCCESS", 25)
addLoggingLevel("HUGEINFO", 21)
addLoggingLevel("HUGEVERBOSE", 16)
addLoggingLevel("VERBOSE", 15)


verbosity_levels_toggle = [logging.INFO, logging.VERBOSE, logging.DEBUG]


def get_bbot_loggers():
    global bbot_loggers
    if bbot_loggers is None:
        bbot_loggers = [
            logging.getLogger("bbot"),
            logging.getLogger("asyncio"),
        ]
    return bbot_loggers


def add_log_handler(handler, formatter=None):
    if handler.formatter is None:
        handler.setFormatter(debug_format)
    for logger in get_bbot_loggers():
        if handler not in logger.handlers:
            logger.addHandler(handler)


def remove_log_handler(handler):
    for logger in get_bbot_loggers():
        if handler in logger.handlers:
            logger.removeHandler(handler)


def init_logging():
    # Don't do this more than once
    if len(logging.getLogger("bbot").handlers) == 0:
        for logger in get_bbot_loggers():
            include_logger(logger)


def include_logger(logger):
    bbot_loggers = get_bbot_loggers()
    if logger not in bbot_loggers:
        bbot_loggers.append(logger)
    logger.setLevel(get_log_level())
    for handler in get_log_handlers().values():
        logger.addHandler(handler)


def get_log_handlers():
    global bbot_log_handlers

    if bbot_log_handlers is None:
        log_dir = Path(config["home"]) / "logs"
        if not mkdir(log_dir, raise_error=False):
            error_and_exit(f"Failure creating or error writing to BBOT logs directory ({log_dir})")

        # Main log file
        main_handler = logging.handlers.TimedRotatingFileHandler(
            f"{log_dir}/bbot.log", when="d", interval=1, backupCount=14
        )

        # Separate log file for debugging
        debug_handler = logging.handlers.TimedRotatingFileHandler(
            f"{log_dir}/bbot.debug.log", when="d", interval=1, backupCount=14
        )

        def stderr_filter(record):
            log_level = get_log_level()
            if record.levelno == logging.STDOUT or (record.levelno == logging.TRACE and log_level > logging.DEBUG):
                return False
            if record.levelno < log_level:
                return False
            return True

        # Log to stderr
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.addFilter(stderr_filter)
        # Log to stdout
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.addFilter(lambda x: x.levelno == logging.STDOUT)
        # log to files
        debug_handler.addFilter(
            lambda x: x.levelno == logging.TRACE or (x.levelno < logging.VERBOSE and x.levelno != logging.STDOUT)
        )
        main_handler.addFilter(
            lambda x: x.levelno not in (logging.STDOUT, logging.TRACE) and x.levelno >= logging.VERBOSE
        )

        # Set log format
        debug_handler.setFormatter(debug_format)
        main_handler.setFormatter(debug_format)
        stderr_handler.setFormatter(ColoredFormatter("%(levelname)s %(name)s: %(message)s"))
        stdout_handler.setFormatter(logging.Formatter("%(message)s"))

        bbot_log_handlers = {
            "stderr": stderr_handler,
            "stdout": stdout_handler,
            "file_debug": debug_handler,
            "file_main": main_handler,
        }
    return bbot_log_handlers


def get_log_level():
    if _log_level_override is not None:
        return _log_level_override

    from bbot.core.configurator.args import cli_options

    if config.get("debug", False) or os.environ.get("BBOT_DEBUG", "").lower() in ("true", "yes"):
        return logging.DEBUG

    loglevel = logging.INFO
    if cli_options is not None:
        if cli_options.verbose:
            loglevel = logging.VERBOSE
        if cli_options.debug:
            loglevel = logging.DEBUG
    return loglevel


def set_log_level(level, logger=None):
    global _log_level_override
    if logger is not None:
        logger.hugeinfo(f"Setting log level to {logging.getLevelName(level)}")
    config["silent"] = False
    _log_level_override = level
    for logger in bbot_loggers:
        logger.setLevel(level)


def toggle_log_level(logger=None):
    log_level = get_log_level()
    if log_level in verbosity_levels_toggle:
        for i, level in enumerate(verbosity_levels_toggle):
            if log_level == level:
                set_log_level(verbosity_levels_toggle[(i + 1) % len(verbosity_levels_toggle)], logger=logger)
    else:
        set_log_level(verbosity_levels_toggle[0], logger=logger)



from ..bbot_fixtures import *


@pytest.mark.asyncio
async def test_python_api(bbot_config):
    from bbot.scanner import Scanner

    # make sure events are properly yielded
    scan1 = Scanner("127.0.0.1", config=bbot_config)
    events1 = []
    async for event in scan1.async_start():
        events1.append(event)
    assert any("127.0.0.1" == e for e in events1)
    # make sure output files work
    scan2 = Scanner("127.0.0.1", config=bbot_config, output_modules=["json"], name="python_api_test")
    await scan2.async_start_without_generator()
    scan_home = scan2.helpers.scans_dir / "python_api_test"
    out_file = scan_home / "output.ndjson"
    assert list(scan2.helpers.read_file(out_file))
    scan_log = scan_home / "scan.log"
    debug_log = scan_home / "debug.log"
    assert scan_log.is_file()
    assert "python_api_test" in open(scan_log).read()
    assert debug_log.is_file()
    assert "python_api_test" in open(debug_log).read()

    scan3 = Scanner("127.0.0.1", config=bbot_config, output_modules=["json"], name="scan_logging_test")
    await scan3.async_start_without_generator()

    assert "scan_logging_test" not in open(scan_log).read()
    assert "scan_logging_test" not in open(debug_log).read()

    scan_home = scan3.helpers.scans_dir / "scan_logging_test"
    out_file = scan_home / "output.ndjson"
    assert list(scan3.helpers.read_file(out_file))
    scan_log = scan_home / "scan.log"
    debug_log = scan_home / "debug.log"
    assert scan_log.is_file()
    assert debug_log.is_file()
    assert "scan_logging_test" in open(scan_log).read()
    assert "scan_logging_test" in open(debug_log).read()

    # make sure config loads properly
    bbot_home = "/tmp/.bbot_python_api_test"
    Scanner("127.0.0.1", config={"home": bbot_home})
    assert os.environ["BBOT_TOOLS"] == str(Path(bbot_home) / "tools")


def test_python_api_sync(bbot_config):
    from bbot.scanner import Scanner

    # make sure events are properly yielded
    scan1 = Scanner("127.0.0.1", config=bbot_config)
    events1 = []
    for event in scan1.start():
        events1.append(event)
    assert any("127.0.0.1" == e for e in events1)
    # make sure output files work
    scan2 = Scanner("127.0.0.1", config=bbot_config, output_modules=["json"], name="python_api_test")
    scan2.start_without_generator()
    out_file = scan2.helpers.scans_dir / "python_api_test" / "output.ndjson"
    assert list(scan2.helpers.read_file(out_file))
    # make sure config loads properly
    bbot_home = "/tmp/.bbot_python_api_test"
    Scanner("127.0.0.1", config={"home": bbot_home})
    assert os.environ["BBOT_TOOLS"] == str(Path(bbot_home) / "tools")



from .base import ModuleTestBase


class TestCensys(ModuleTestBase):
    config_overrides = {"modules": {"censys": {"api_id": "api_id", "api_secret": "api_secret"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://search.censys.io/api/v1/account",
            json={
                "email": "info@blacklanternsecurity.com",
                "login": "nope",
                "first_login": "1917-08-03 20:03:55",
                "last_login": "1918-05-19 01:15:22",
                "quota": {"used": 26, "allowance": 250, "resets_at": "1919-06-03 16:30:32"},
            },
        )
        module_test.httpx_mock.add_response(
            url="https://search.censys.io/api/v2/certificates/search",
            match_content=b'{"q": "names: blacklanternsecurity.com", "per_page": 100}',
            json={
                "code": 200,
                "status": "OK",
                "result": {
                    "query": "names: blacklanternsecurity.com",
                    "total": 196,
                    "duration_ms": 1046,
                    "hits": [
                        {
                            "parsed": {
                                "validity_period": {
                                    "not_before": "2021-11-18T00:09:46Z",
                                    "not_after": "2022-11-18T00:09:46Z",
                                },
                                "issuer_dn": "C=US, ST=Arizona, L=Scottsdale, O=GoDaddy.com\\, Inc., OU=http://certs.godaddy.com/repository/, CN=Go Daddy Secure Certificate Authority - G2",
                                "subject_dn": "CN=asdf.blacklanternsecurity.com",
                            },
                            "fingerprint_sha256": "590ad51b8db62925f0fd3f300264c6a36692e20ceec2b5a22e7e4b41c1575cdc",
                            "names": ["asdf.blacklanternsecurity.com", "asdf2.blacklanternsecurity.com"],
                        },
                    ],
                    "links": {"next": "NextToken", "prev": ""},
                },
            },
        )
        module_test.httpx_mock.add_response(
            url="https://search.censys.io/api/v2/certificates/search",
            match_content=b'{"q": "names: blacklanternsecurity.com", "per_page": 100, "cursor": "NextToken"}',
            json={
                "code": 200,
                "status": "OK",
                "result": {
                    "query": "names: blacklanternsecurity.com",
                    "total": 196,
                    "duration_ms": 1046,
                    "hits": [
                        {
                            "parsed": {
                                "validity_period": {
                                    "not_before": "2021-11-18T00:09:46Z",
                                    "not_after": "2022-11-18T00:09:46Z",
                                },
                                "issuer_dn": "C=US, ST=Arizona, L=Scottsdale, O=GoDaddy.com\\, Inc., OU=http://certs.godaddy.com/repository/, CN=Go Daddy Secure Certificate Authority - G2",
                                "subject_dn": "CN=zzzz.blacklanternsecurity.com",
                            },
                            "fingerprint_sha256": "590ad51b8db62925f0fd3f300264c6a36692e20ceec2b5a22e7e4b41c1575cdc",
                            "names": ["zzzz.blacklanternsecurity.com"],
                        },
                    ],
                    "links": {"next": "", "prev": ""},
                },
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect asdf subdomain"
        assert any(e.data == "asdf2.blacklanternsecurity.com" for e in events), "Failed to detect asdf2 subdomain"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect zzzz subdomain"



from .base import ModuleTestBase


class TestNSEC(ModuleTestBase):
    targets = ["blacklanternsecurity.notreal"]

    async def setup_after_prep(self, module_test):
        next(iter(module_test.scan.target.events)).add_tag("ns-record")

        old_resolve_fn = module_test.scan.helpers.dns.resolve

        async def resolve(query, **kwargs):
            if query == "blacklanternsecurity.notreal" and kwargs.get("type", "").upper() == "NSEC":
                return {"asdf.blacklanternsecurity.notreal"}
            elif query == "asdf.blacklanternsecurity.notreal" and kwargs.get("type", "").upper() == "NSEC":
                return {"zzzz.blacklanternsecurity.notreal"}
            elif query == "zzzz.blacklanternsecurity.notreal" and kwargs.get("type", "").upper() == "NSEC":
                return {"blacklanternsecurity.notreal"}
            return await old_resolve_fn(query, **kwargs)

        module_test.monkeypatch.setattr(module_test.scan.helpers.dns, "resolve", resolve)

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.notreal" for e in events), "Failed to detect subdomain #1"
        assert any(e.data == "zzzz.blacklanternsecurity.notreal" for e in events), "Failed to detect subdomain #2"



import os
import logging
from pathlib import Path

from . import misc
from .dns import DNSHelper
from .web import WebHelper
from .diff import HttpCompare
from .cloud import CloudHelper
from .wordcloud import WordCloud
from .interactsh import Interactsh
from ...scanner.target import Target
from ...modules.base import BaseModule
from .depsinstaller import DepsInstaller


log = logging.getLogger("bbot.core.helpers")


class ConfigAwareHelper:
    """
    Centralized helper class that provides unified access to various helper functions.

    This class serves as a convenient interface for accessing helper methods across different files.
    It is designed to be configuration-aware, allowing helper functions to utilize scan-specific
    configurations like rate-limits. The class leverages Python's `__getattribute__` magic method
    to provide seamless access to helper functions across various namespaces.

    Attributes:
        config (dict): Configuration settings for the BBOT scan instance.
        _scan (Scan): A BBOT scan instance.
        bbot_home (Path): Home directory for BBOT.
        cache_dir (Path): Directory for storing cache files.
        temp_dir (Path): Directory for storing temporary files.
        tools_dir (Path): Directory for storing tools, e.g. compiled binaries.
        lib_dir (Path): Directory for storing libraries.
        scans_dir (Path): Directory for storing scan results.
        wordlist_dir (Path): Directory for storing wordlists.
        current_dir (Path): The current working directory.
        keep_old_scans (int): The number of old scans to keep.

    Examples:
        >>> helper = ConfigAwareHelper(config)
        >>> ips = helper.dns.resolve("www.evilcorp.com")
    """

    from . import ntlm
    from . import regexes
    from . import validators
    from .files import tempfile, feed_pipe, _feed_pipe, tempfile_tail
    from .command import run, run_live, _spawn_proc, _prepare_command_kwargs
    from .cache import cache_get, cache_put, cache_filename, is_cached, CacheDict

    def __init__(self, config, scan=None):
        self.config = config
        self._scan = scan
        self.bbot_home = Path(self.config.get("home", "~/.bbot")).expanduser().resolve()
        self.cache_dir = self.bbot_home / "cache"
        self.temp_dir = self.bbot_home / "temp"
        self.tools_dir = self.bbot_home / "tools"
        self.lib_dir = self.bbot_home / "lib"
        self.scans_dir = self.bbot_home / "scans"
        self.wordlist_dir = Path(__file__).parent.parent.parent / "wordlists"
        self.current_dir = Path.cwd()
        self.keep_old_scans = self.config.get("keep_scans", 20)
        self.mkdir(self.cache_dir)
        self.mkdir(self.temp_dir)
        self.mkdir(self.tools_dir)
        self.mkdir(self.lib_dir)

        self.dns = DNSHelper(self)
        self.web = WebHelper(self)
        self.depsinstaller = DepsInstaller(self)
        self.word_cloud = WordCloud(self)
        self.dummy_modules = {}

        # cloud helpers
        self.cloud = CloudHelper(self)

    def interactsh(self, *args, **kwargs):
        return Interactsh(self, *args, **kwargs)

    def http_compare(self, url, allow_redirects=False, include_cache_buster=True):
        return HttpCompare(url, self, allow_redirects=allow_redirects, include_cache_buster=include_cache_buster)

    def temp_filename(self, extension=None):
        """
        temp_filename() --> Path("/home/user/.bbot/temp/pgxml13bov87oqrvjz7a")
        """
        filename = self.rand_string(20)
        if extension is not None:
            filename = f"{filename}.{extension}"
        return self.temp_dir / filename

    def clean_old_scans(self):
        _filter = lambda x: x.is_dir() and self.regexes.scan_name_regex.match(x.name)
        self.clean_old(self.scans_dir, keep=self.keep_old_scans, filter=_filter)

    def make_target(self, *events):
        return Target(self.scan, *events)

    @property
    def scan(self):
        if self._scan is None:
            from bbot.scanner import Scanner

            self._scan = Scanner()
        return self._scan

    @property
    def in_tests(self):
        return os.environ.get("BBOT_TESTING", "") == "True"

    def _make_dummy_module(self, name, _type="scan"):
        """
        Construct a dummy module, for attachment to events
        """
        try:
            return self.dummy_modules[name]
        except KeyError:
            dummy = DummyModule(scan=self.scan, name=name, _type=_type)
            self.dummy_modules[name] = dummy
            return dummy

    def __getattribute__(self, attr):
        """
        Do not be afraid, the angel said.

        Overrides Python's built-in __getattribute__ to provide convenient access to helper methods.

        This method first attempts to find an attribute within this class itself. If unsuccessful,
        it then looks in the 'misc', 'dns', and 'web' helper modules, in that order. If the attribute
        is still not found, an AttributeError is raised.

        Args:
            attr (str): The attribute name to look for.

        Returns:
            Any: The attribute value, if found.

        Raises:
            AttributeError: If the attribute is not found in any of the specified places.
        """
        try:
            # first try self
            return super().__getattribute__(attr)
        except AttributeError:
            try:
                # then try misc
                return getattr(misc, attr)
            except AttributeError:
                try:
                    # then try dns
                    return getattr(self.dns, attr)
                except AttributeError:
                    try:
                        # then try web
                        return getattr(self.web, attr)
                    except AttributeError:
                        try:
                            # then try validators
                            return getattr(self.validators, attr)
                        except AttributeError:
                            # then die
                            raise AttributeError(f'Helper has no attribute "{attr}"')


class DummyModule(BaseModule):
    _priority = 4

    def __init__(self, *args, **kwargs):
        self._name = kwargs.pop("name")
        self._type = kwargs.pop("_type")
        super().__init__(*args, **kwargs)



import re
import json
import random
import subprocess

from bbot.modules.templates.subdomain_enum import subdomain_enum


class massdns(subdomain_enum):
    """
    This is BBOT's flagship subdomain enumeration module.

    It uses massdns to brute-force subdomains.
    At the end of a scan, it will leverage BBOT's word cloud to recursively discover target-specific subdomain mutations.

    Each subdomain discovered via mutations is tagged with the "mutation" tag. This tag includes the depth at which
    the mutations is found. I.e. the first mutation will be tagged "mutation-1". The second one (a mutation of a
    mutation) will be "mutation-2". Mutations of mutations of mutations will be "mutation-3", etc.

    This is especially use for bug bounties because it enables you to recognize distant/rare subdomains at a glance.
    Subdomains with higher mutation levels are more likely to be distant/rare or never-before-seen.
    """

    flags = ["subdomain-enum", "passive", "aggressive"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Brute-force subdomains with massdns (highly effective)"}
    options = {
        "wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/master/Discovery/DNS/subdomains-top1million-5000.txt",
        "max_resolvers": 1000,
        "max_mutations": 500,
    }
    options_desc = {
        "wordlist": "Subdomain wordlist URL",
        "max_resolvers": "Number of concurrent massdns resolvers",
        "max_mutations": "Max number of smart mutations per subdomain",
    }
    subdomain_file = None
    deps_ansible = [
        {
            "name": "install dev tools",
            "package": {"name": ["gcc", "git", "make"], "state": "present"},
            "become": True,
            "ignore_errors": True,
        },
        {
            "name": "Download massdns source code",
            "git": {
                "repo": "https://github.com/blechschmidt/massdns.git",
                "dest": "#{BBOT_TEMP}/massdns",
                "single_branch": True,
                "version": "master",
            },
        },
        {
            "name": "Build massdns (Linux)",
            "command": {"chdir": "#{BBOT_TEMP}/massdns", "cmd": "make", "creates": "#{BBOT_TEMP}/massdns/bin/massdns"},
            "when": "ansible_facts['system'] == 'Linux'",
        },
        {
            "name": "Build massdns (non-Linux)",
            "command": {
                "chdir": "#{BBOT_TEMP}/massdns",
                "cmd": "make nolinux",
                "creates": "#{BBOT_TEMP}/massdns/bin/massdns",
            },
            "when": "ansible_facts['system'] != 'Linux'",
        },
        {
            "name": "Install massdns",
            "copy": {"src": "#{BBOT_TEMP}/massdns/bin/massdns", "dest": "#{BBOT_TOOLS}/", "mode": "u+x,g+x,o+x"},
        },
    ]
    reject_wildcards = "strict"
    _qsize = 100

    digit_regex = re.compile(r"\d+")

    async def setup(self):
        self.found = dict()
        self.mutations_tried = set()
        self.source_events = self.helpers.make_target()
        self.subdomain_file = await self.helpers.wordlist(self.config.get("wordlist"))
        self.subdomain_list = set(self.helpers.read_file(self.subdomain_file))

        ms_on_prem_string_file = self.helpers.wordlist_dir / "ms_on_prem_subdomains.txt"
        ms_on_prem_strings = set(self.helpers.read_file(ms_on_prem_string_file))
        self.subdomain_list.update(ms_on_prem_strings)

        self.max_resolvers = self.config.get("max_resolvers", 1000)
        self.max_mutations = self.config.get("max_mutations", 500)
        nameservers_url = (
            "https://raw.githubusercontent.com/blacklanternsecurity/public-dns-servers/master/nameservers.txt"
        )
        self.resolver_file = await self.helpers.wordlist(
            nameservers_url,
            cache_hrs=24 * 7,
        )
        self.devops_mutations = list(self.helpers.word_cloud.devops_mutations)
        self._mutation_run = 1
        return await super().setup()

    async def filter_event(self, event):
        query = self.make_query(event)
        eligible, reason = await self.eligible_for_enumeration(event)
        if eligible:
            self.add_found(event)
        # reject if already processed
        if self.already_processed(query):
            return False, f'Query "{query}" was already processed'
        if eligible:
            self.processed.add(hash(query))
            return True, reason
        return False, reason

    async def handle_event(self, event):
        query = self.make_query(event)
        self.source_events.add_target(event)

        self.info(f"Brute-forcing subdomains for {query} (source: {event.data})")
        for hostname in await self.massdns(query, self.subdomain_list):
            await self.emit_result(hostname, event, query)

    def abort_if(self, event):
        if not event.scope_distance == 0:
            return True, "event is not in scope"
        if "wildcard" in event.tags:
            return True, "event is a wildcard"

    async def emit_result(self, result, source_event, query, tags=None):
        if not result == source_event:
            kwargs = {"abort_if": self.abort_if}
            if tags is not None:
                kwargs["tags"] = tags
            await self.emit_event(result, "DNS_NAME", source_event, **kwargs)

    def already_processed(self, hostname):
        if hash(hostname) in self.processed:
            return True
        return False

    async def massdns(self, domain, subdomains):
        subdomains = list(subdomains)

        domain_wildcard_rdtypes = set()
        for domain, rdtypes in (await self.helpers.is_wildcard_domain(domain)).items():
            for rdtype, results in rdtypes.items():
                if results:
                    domain_wildcard_rdtypes.add(rdtype)
        if any([r in domain_wildcard_rdtypes for r in ("A", "CNAME")]):
            self.info(
                f"Aborting massdns on {domain} because it's a wildcard domain ({','.join(domain_wildcard_rdtypes)})"
            )
            self.found.pop(domain, None)
            return []
        else:
            self.log.trace(f"{domain}: A is not in domain_wildcard_rdtypes:{domain_wildcard_rdtypes}")

        # before we start, do a canary check for wildcards
        abort_msg = f"Aborting massdns on {domain} due to false positive"
        canary_result = await self._canary_check(domain)
        if canary_result:
            self.info(abort_msg + f": {canary_result}")
            return []
        else:
            self.log.trace(f"Canary result for {domain}: {canary_result}")

        results = []
        async for hostname, ip, rdtype in self._massdns(domain, subdomains):
            # allow brute-forcing of wildcard domains
            # this is dead code but it's kinda cool so it can live here
            if rdtype in domain_wildcard_rdtypes:
                # skip wildcard checking on multi-level subdomains for performance reasons
                stem = hostname.split(domain)[0].strip(".")
                if "." in stem:
                    self.debug(f"Skipping {hostname}:A because it may be a wildcard (reason: performance)")
                    continue
                wildcard_rdtypes = await self.helpers.is_wildcard(hostname, ips=(ip,), rdtype=rdtype)
                if rdtype in wildcard_rdtypes:
                    self.debug(f"Skipping {hostname}:{rdtype} because it's a wildcard")
                    continue
            results.append(hostname)

        # do another canary check for good measure
        if len(results) > 50:
            canary_result = await self._canary_check(domain)
            if canary_result:
                self.info(abort_msg + f": {canary_result}")
                return []
            else:
                self.log.trace(f"Canary result for {domain}: {canary_result}")

        # abort if there are a suspiciously high number of results
        # (the results are over 2000, and this is more than 20 percent of the input size)
        if len(results) > 2000:
            if len(results) / len(subdomains) > 0.2:
                self.info(
                    f"Aborting because the number of results ({len(results):,}) is suspiciously high for the length of the wordlist ({len(subdomains):,})"
                )
                return []
            else:
                self.info(
                    f"{len(results):,} results returned from massdns against {domain} (wordlist size = {len(subdomains):,})"
                )

        # everything checks out
        self.verbose(f"Resolving batch of {len(results):,} results")
        resolved = dict([l async for l in self.helpers.resolve_batch(results, type=("A", "CNAME"))])
        resolved = {k: v for k, v in resolved.items() if v}
        for hostname in resolved:
            self.add_found(hostname)
        return list(resolved)

    async def _canary_check(self, domain, num_checks=50):
        random_subdomains = list(self.gen_random_subdomains(num_checks))
        self.verbose(f"Testing {len(random_subdomains):,} canaries against {domain}")
        canary_results = [h async for h, d, r in self._massdns(domain, random_subdomains)]
        self.log.trace(f"canary results for {domain}: {canary_results}")
        resolved_canaries = self.helpers.resolve_batch(canary_results)
        self.log.trace(f"resolved canary results for {domain}: {canary_results}")
        async for query, result in resolved_canaries:
            if result:
                await resolved_canaries.aclose()
                result = f"{query}:{result}"
                self.log.trace(f"Found false positive: {result}")
                return result
        self.log.trace(f"Passed canary check for {domain}")
        return False

    async def _massdns(self, domain, subdomains):
        """
        {
          "name": "www.blacklanternsecurity.com.",
          "type": "A",
          "class": "IN",
          "status": "NOERROR",
          "data": {
            "answers": [
              {
                "ttl": 3600,
                "type": "CNAME",
                "class": "IN",
                "name": "www.blacklanternsecurity.com.",
                "data": "blacklanternsecurity.github.io."
              },
              {
                "ttl": 3600,
                "type": "A",
                "class": "IN",
                "name": "blacklanternsecurity.github.io.",
                "data": "185.199.108.153"
              }
            ]
          },
          "resolver": "168.215.165.186:53"
        }
        """
        if self.scan.stopping:
            return

        command = (
            "massdns",
            "-r",
            self.resolver_file,
            "-s",
            self.max_resolvers,
            "-t",
            "A",
            "-o",
            "J",
            "-q",
        )
        subdomains = self.gen_subdomains(subdomains, domain)
        hosts_yielded = set()
        async for line in self.helpers.run_live(command, stderr=subprocess.DEVNULL, input=subdomains):
            try:
                j = json.loads(line)
            except json.decoder.JSONDecodeError:
                self.debug(f"Failed to decode line: {line}")
                continue
            answers = j.get("data", {}).get("answers", [])
            if type(answers) == list and len(answers) > 0:
                answer = answers[0]
                hostname = answer.get("name", "").strip(".").lower()
                if hostname.endswith(f".{domain}"):
                    data = answer.get("data", "")
                    rdtype = answer.get("type", "").upper()
                    # avoid garbage answers like this:
                    # 8AAAA queries have been locally blocked by dnscrypt-proxy/Set block_ipv6 to false to disable this feature
                    if data and rdtype and not " " in data:
                        hostname_hash = hash(hostname)
                        if hostname_hash not in hosts_yielded:
                            hosts_yielded.add(hostname_hash)
                            yield hostname, data, rdtype

    async def finish(self):
        found = sorted(self.found.items(), key=lambda x: len(x[-1]), reverse=True)
        # if we have a lot of rounds to make, don't try mutations on less-populated domains
        trimmed_found = []
        if found:
            avg_subdomains = sum([len(subdomains) for domain, subdomains in found[:50]]) / len(found[:50])
            for i, (domain, subdomains) in enumerate(found):
                # accept domains that are in the top 50 or have more than 5 percent of the average number of subdomains
                if i < 50 or (len(subdomains) > 1 and len(subdomains) >= (avg_subdomains * 0.05)):
                    trimmed_found.append((domain, subdomains))
                else:
                    self.verbose(
                        f"Skipping mutations on {domain} because it only has {len(subdomains):,} subdomain(s) (avg: {avg_subdomains:,})"
                    )

        base_mutations = set()
        found_mutations = False
        try:
            for i, (domain, subdomains) in enumerate(trimmed_found):
                self.verbose(f"{domain} has {len(subdomains):,} subdomains")
                # keep looping as long as we're finding things
                while 1:
                    max_mem_percent = 90
                    mem_status = self.helpers.memory_status()
                    # abort if we don't have the memory
                    mem_percent = mem_status.percent
                    if mem_percent > max_mem_percent:
                        free_memory = mem_status.available
                        free_memory_human = self.helpers.bytes_to_human(free_memory)
                        assert (
                            False
                        ), f"Cannot proceed with DNS mutations because system memory is at {mem_percent:.1f}% ({free_memory_human} remaining)"

                    query = domain
                    domain_hash = hash(domain)
                    if self.scan.stopping:
                        return

                    mutations = set(base_mutations)

                    def add_mutation(_domain_hash, m):
                        h = hash((_domain_hash, m))
                        if h not in self.mutations_tried:
                            self.mutations_tried.add(h)
                            mutations.add(m)

                    # try every subdomain everywhere else
                    for _domain, _subdomains in found:
                        if _domain == domain:
                            continue
                        for s in _subdomains:
                            first_segment = s.split(".")[0]
                            # skip stuff with lots of numbers (e.g. PTRs)
                            digits = self.digit_regex.findall(first_segment)
                            excessive_digits = len(digits) > 2
                            long_digits = any(len(d) > 3 for d in digits)
                            if excessive_digits or long_digits:
                                continue
                            add_mutation(domain_hash, first_segment)
                            for word in self.helpers.extract_words(
                                first_segment, word_regexes=self.helpers.word_cloud.dns_mutator.extract_word_regexes
                            ):
                                add_mutation(domain_hash, word)

                    # numbers + devops mutations
                    for mutation in self.helpers.word_cloud.mutations(
                        subdomains, cloud=False, numbers=3, number_padding=1
                    ):
                        for delimiter in ("", ".", "-"):
                            m = delimiter.join(mutation).lower()
                            add_mutation(domain_hash, m)

                    # special dns mutator
                    for subdomain in self.helpers.word_cloud.dns_mutator.mutations(
                        subdomains, max_mutations=self.max_mutations
                    ):
                        add_mutation(domain_hash, subdomain)

                    if mutations:
                        self.info(f"Trying {len(mutations):,} mutations against {domain} ({i+1}/{len(found)})")
                        results = list(await self.massdns(query, mutations))
                        for hostname in results:
                            source_event = self.source_events.get(hostname)
                            if source_event is None:
                                self.warning(f"Could not correlate source event from: {hostname}")
                                source_event = self.scan.root_event
                            await self.emit_result(
                                hostname, source_event, query, tags=[f"mutation-{self._mutation_run}"]
                            )
                        if results:
                            found_mutations = True
                            continue
                    break
        except AssertionError as e:
            self.warning(e)

        if found_mutations:
            self._mutation_run += 1

    def add_found(self, host):
        if not isinstance(host, str):
            host = host.data
        if self.helpers.is_subdomain(host):
            subdomain, domain = host.split(".", 1)
            is_ptr = self.helpers.is_ptr(subdomain)
            in_scope = self.scan.in_scope(domain)
            if in_scope and not is_ptr:
                try:
                    self.found[domain].add(subdomain)
                except KeyError:
                    self.found[domain] = set((subdomain,))

    async def gen_subdomains(self, prefixes, domain):
        for p in prefixes:
            d = f"{p}.{domain}"
            yield d

    def gen_random_subdomains(self, n=50):
        delimiters = (".", "-")
        lengths = list(range(3, 8))
        for i in range(0, max(0, n - 5)):
            d = delimiters[i % len(delimiters)]
            l = lengths[i % len(lengths)]
            segments = list(random.choice(self.devops_mutations) for _ in range(l))
            segments.append(self.helpers.rand_string(length=8, digits=False))
            subdomain = d.join(segments)
            yield subdomain
        for _ in range(5):
            yield self.helpers.rand_string(length=8, digits=False)



from lxml import etree
from bbot.modules.templates.portscanner import portscanner


class nmap(portscanner):
    watched_events = ["IP_ADDRESS", "DNS_NAME", "IP_RANGE"]
    produced_events = ["OPEN_TCP_PORT"]
    flags = ["active", "portscan", "aggressive", "web-thorough"]
    meta = {"description": "Port scan with nmap. By default, scans top 100 ports."}
    options = {
        "top_ports": 100,
        "ports": "",
        "timing": "T4",
        "skip_host_discovery": True,
    }
    options_desc = {
        "top_ports": "Top ports to scan (default 100) (to override, specify 'ports')",
        "ports": "Ports to scan",
        "timing": "-T<0-5>: Set timing template (higher is faster)",
        "skip_host_discovery": "skip host discovery (-Pn)",
    }
    _max_event_handlers = 2
    batch_size = 256
    _priority = 2

    deps_apt = ["nmap"]
    deps_pip = ["lxml~=4.9.2"]

    async def setup(self):
        self.helpers.depsinstaller.ensure_root(message="Nmap requires root privileges")
        self.ports = self.config.get("ports", "")
        self.timing = self.config.get("timing", "T4")
        self.top_ports = self.config.get("top_ports", 100)
        self.skip_host_discovery = self.config.get("skip_host_discovery", True)
        return await super().setup()

    async def handle_batch(self, *events):
        target = self.helpers.make_target(*events)
        targets = list(set(str(e.data) for e in events))
        command, output_file = self.construct_command(targets)
        try:
            await self.helpers.run(command, sudo=True)
            for host in self.parse_nmap_xml(output_file):
                source_event = None
                for h in [host.address] + host.hostnames:
                    source_event = target.get(h)
                    if source_event is not None:
                        break
                if source_event is None:
                    self.warning(f"Failed to correlate source event from {host}")
                    source_event = self.scan.root_event
                for port in host.open_ports:
                    port_number = int(port.split("/")[0])
                    netloc = self.helpers.make_netloc(host.address, port_number)
                    await self.emit_event(netloc, "OPEN_TCP_PORT", source=source_event)
                    for hostname in host.hostnames:
                        netloc = self.helpers.make_netloc(hostname, port_number)
                        await self.emit_event(netloc, "OPEN_TCP_PORT", source=source_event)
        finally:
            output_file.unlink(missing_ok=True)

    def construct_command(self, targets):
        ports = self.config.get("ports", "")
        top_ports = self.config.get("top_ports", "")
        temp_filename = self.helpers.temp_filename(extension="xml")
        command = [
            "nmap",
            "--noninteractive",
            "--excludefile",
            str(self.exclude_file),
            "-n",
            "--resolve-all",
            f"-{self.timing}",
            "-oX",
            temp_filename,
        ]
        if self.skip_host_discovery:
            command += ["-Pn"]
        if ports:
            command += ["-p", ports]
        else:
            command += ["--top-ports", top_ports]
        command += targets
        return command, temp_filename

    def parse_nmap_xml(self, xml_file):
        try:
            with open(xml_file, "rb") as f:
                et = etree.parse(f)
                for host in et.iter("host"):
                    yield NmapHost(host)
        except Exception as e:
            self.warning(f"Error parsing Nmap XML at {xml_file}: {e}")

    async def cleanup(self):
        resume_file = self.helpers.current_dir / "resume.cfg"
        resume_file.unlink(missing_ok=True)


class NmapHost(str):
    def __init__(self, xml):
        self.etree = xml

        # convenient host information
        self.status = self.etree.find("status").attrib.get("state", "down")
        self.address = self.etree.find("address").attrib.get("addr", "")
        self.hostnames = []
        for hostname in self.etree.findall("hostnames/hostname"):
            hostname = hostname.attrib.get("name")
            if hostname and not hostname in self.hostnames:
                self.hostnames.append(hostname)

        # convenient port information
        self.scripts = dict()
        self.open_ports = []
        self.closed_ports = []
        self.filtered_ports = []
        for port in self.etree.findall("ports/port"):
            port_name = port.attrib.get("portid", "0") + "/" + port.attrib.get("protocol", "tcp").lower()
            port_status = port.find("state").attrib.get("state", "closed")
            if port_status in ("open", "closed", "filtered"):
                getattr(self, f"{port_status}_ports").append(port_name)
            for script in port.iter("script"):
                script_name = script.attrib.get("id", "")
                script_output = script.attrib.get("output", "")
                if script_name:
                    try:
                        self.scripts[port_name][script_name] = script_output
                    except KeyError:
                        self.scripts[port_name] = {script_name: script_output}

    def __str__(self):
        address = self.address + (" " if self.address else "")
        hostnames = "(" + ", ".join(self.hostnames) + ")" if self.hostnames else ""
        return f"{address}{hostnames}"

    def __repr__(self):
        return str(self)



from bbot.modules.templates.subdomain_enum import subdomain_enum


class threatminer(subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query threatminer's API for subdomains",
    }

    base_url = "https://api.threatminer.org/v2"

    async def request_url(self, query):
        url = f"{self.base_url}/domain.php?q={self.helpers.quote(query)}&rt=5"
        r = await self.request_with_fail_count(url, timeout=self.http_timeout + 30)
        return r

    def parse_results(self, r, query):
        j = r.json()
        return list(j.get("results", []))



import json
from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class JSON(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to Newline-Delimited JSON (NDJSON)"}
    options = {"output_file": "", "console": False}
    options_desc = {"output_file": "Output to file", "console": "Output to console"}
    _preserve_graph = True

    async def setup(self):
        self._prep_output_dir("output.ndjson")
        return True

    async def handle_event(self, event):
        event_str = json.dumps(dict(event))
        if self.file is not None:
            self.file.write(event_str + "\n")
            self.file.flush()
        if self.config.get("console", False) or "human" not in self.scan.modules:
            self.stdout(event_str)

    async def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    async def report(self):
        if self._file is not None:
            self.info(f"Saved JSON output to {self.output_file}")



from bbot.modules.base import BaseModule


class internetdb(BaseModule):
    """
    Query IP in Shodan InternetDB, returning open ports, discovered technologies, and findings/vulnerabilities
    API reference: https://internetdb.shodan.io/docs

    Example API response:

    {
        "cpes": [
            "cpe:/a:microsoft:internet_information_services",
            "cpe:/a:microsoft:outlook_web_access:15.0.1367",
        ],
        "hostnames": [
            "autodiscover.evilcorp.com",
            "mail.evilcorp.com",
        ],
        "ip": "1.2.3.4",
        "ports": [
            25,
            80,
            443,
        ],
        "tags": [
            "starttls",
            "self-signed",
            "eol-os"
        ],
        "vulns": [
            "CVE-2021-26857",
            "CVE-2021-26855"
        ]
    }
    """

    watched_events = ["IP_ADDRESS", "DNS_NAME"]
    produced_events = ["TECHNOLOGY", "VULNERABILITY", "FINDING", "OPEN_TCP_PORT", "DNS_NAME"]
    flags = ["passive", "safe", "portscan", "subdomain-enum"]
    meta = {"description": "Query Shodan's InternetDB for open ports, hostnames, technologies, and vulnerabilities"}

    # limit outgoing queue size to help avoid rate limiting
    _qsize = 100

    base_url = "https://internetdb.shodan.io"

    async def setup(self):
        self.processed = set()
        return True

    async def filter_event(self, event):
        ip = self.get_ip(event)
        if ip:
            ip_hash = hash(ip)
            if ip_hash in self.processed:
                return False, "IP was already processed"
            self.processed.add(ip_hash)
            return True
        return False, "event had no valid IP addresses"

    async def handle_event(self, event):
        ip = self.get_ip(event)
        if ip is None:
            return
        url = f"{self.base_url}/{ip}"
        r = await self.request_with_fail_count(url)
        if r is None:
            self.debug(f"No response for {event.data}")
            return
        try:
            data = r.json()
        except Exception as e:
            self.verbose(f"Error parsing JSON response from {url}: {e}")
            self.trace()
            return
        if data:
            if r.status_code == 200:
                await self._parse_response(data=data, event=event)
            elif r.status_code == 404:
                detail = data.get("detail", "")
                if detail:
                    self.debug(f"404 response for {url}: {detail}")
            else:
                err_data = data.get("type", "")
                err_msg = data.get("msg", "")
                self.verbose(f"Shodan error for {ip}: {err_data}: {err_msg}")

    async def _parse_response(self, data: dict, event):
        """Handles emitting events from returned JSON"""
        data: dict  # has keys: cpes, hostnames, ip, ports, tags, vulns
        # ip is a string, ports is a list of ports, the rest is a list of strings
        for hostname in data.get("hostnames", []):
            await self.emit_event(hostname, "DNS_NAME", source=event)
        for cpe in data.get("cpes", []):
            await self.emit_event({"technology": cpe, "host": str(event.host)}, "TECHNOLOGY", source=event)
        for port in data.get("ports", []):
            await self.emit_event(
                self.helpers.make_netloc(event.data, port), "OPEN_TCP_PORT", source=event, internal=True, quick=True
            )
        vulns = data.get("vulns", [])
        if vulns:
            vulns_str = ", ".join([str(v) for v in vulns])
            await self.emit_event(
                {"description": f"Shodan reported verified vulnerabilities: {vulns_str}", "host": str(event.host)},
                "FINDING",
                source=event,
            )

    def get_ip(self, event):
        """
        Get the first available IP address from an event (IP_ADDRESS or DNS_NAME)
        """
        if event.type == "IP_ADDRESS":
            return event.host
        elif event.type == "DNS_NAME":
            # always try IPv4 first
            ipv6 = []
            for host in event.resolved_hosts:
                if self.helpers.is_ip(host, version=4):
                    return host
                elif self.helpers.is_ip(host, version=6):
                    ipv6.append(host)
            for ip in ipv6:
                return ip



from bbot.modules.base import BaseModule
from bbot.core.errors import InteractshError


class host_header(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "web-thorough"]
    meta = {"description": "Try common HTTP Host header spoofing techniques"}

    in_scope_only = True
    per_hostport_only = True

    deps_apt = ["curl"]

    async def setup(self):
        self.subdomain_tags = {}
        if self.scan.config.get("interactsh_disable", False) == False:
            try:
                self.interactsh_instance = self.helpers.interactsh()
                self.domain = await self.interactsh_instance.register(callback=self.interactsh_callback)
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")
                return False
        else:
            self.warning("Interactsh is disabled globally. Interaction based detections will be disabled.")
            self.domain = f"{self.rand_string(12, digits=False)}.com"
        return True

    def rand_string(self, *args, **kwargs):
        return self.helpers.rand_string(*args, **kwargs)

    async def interactsh_callback(self, r):
        full_id = r.get("full-id", None)
        if full_id:
            if "." in full_id:
                match = self.subdomain_tags.get(full_id.split(".")[0])
                if match is None:
                    return
                matched_event = match[0]
                matched_technique = match[1]

                await self.emit_event(
                    {
                        "host": str(matched_event.host),
                        "url": matched_event.data["url"],
                        "description": f"Spoofed Host header ({matched_technique}) [{r.get('protocol').upper()}] interaction",
                    },
                    "FINDING",
                    matched_event,
                )
            else:
                # this is likely caused by something trying to resolve the base domain first and can be ignored
                self.debug("skipping results because subdomain tag was missing")

    async def finish(self):
        if self.scan.config.get("interactsh_disable", False) == False:
            await self.helpers.sleep(5)
            try:
                for r in await self.interactsh_instance.poll():
                    await self.interactsh_callback(r)
            except InteractshError as e:
                self.debug(f"Error in interact.sh: {e}")

    async def cleanup(self):
        if self.scan.config.get("interactsh_disable", False) == False:
            try:
                await self.interactsh_instance.deregister()
                self.debug(
                    f"successfully deregistered interactsh session with correlation_id {self.interactsh_instance.correlation_id}"
                )
            except InteractshError as e:
                self.warning(f"Interactsh failure: {e}")

    async def handle_event(self, event):
        # get any set-cookie responses from the response and add them to the request

        added_cookies = {}

        for header, header_value in event.data["header-dict"].items():
            if header_value.lower() == "set-cookie":
                header_split = header_value.split("=")
                try:
                    added_cookies = {header_split[0]: header_split[1]}
                except IndexError:
                    self.debug(f"failed to parse cookie from string {header_value}")

        domain_reflections = []

        # host header replacement
        technique_description = "standard"
        self.debug(f"Performing {technique_description} case")
        subdomain_tag = self.rand_string(4, digits=False)
        self.subdomain_tags[subdomain_tag] = (event, technique_description)
        output = await self.helpers.curl(
            url=event.data["url"],
            headers={"Host": f"{subdomain_tag}.{self.domain}"},
            ignore_bbot_global_settings=True,
            cookies=added_cookies,
        )
        if self.domain in output:
            domain_reflections.append(technique_description)

        # absolute URL / Host header transposition
        technique_description = "absolute URL transposition"
        self.debug(f"Performing {technique_description} case")
        subdomain_tag = self.rand_string(4, digits=False)
        self.subdomain_tags[subdomain_tag] = (event, technique_description)
        output = await self.helpers.curl(
            url=event.data["url"],
            path_override=event.data["url"],
            cookies=added_cookies,
        )

        if self.domain in output:
            domain_reflections.append(technique_description)

        # duplicate host header tolerance
        technique_description = "duplicate host header tolerance"
        output = await self.helpers.curl(
            url=event.data["url"],
            # Sending a blank HOST first as a hack to trick curl. This makes it no longer an "internal header", thereby allowing for duplicates
            # The fact that it's accepting two host headers is rare enough to note on its own, and not too noisy. Having the 3rd header be an interactsh would result in false negatives for the slightly less interesting cases.
            headers={"Host": ["", str(event.host), str(event.host)]},
            cookies=added_cookies,
            head_mode=True,
        )

        split_output = output.split("\n")
        if " 4" in split_output:
            await self.emit_event(
                {
                    "host": str(event.host),
                    "url": event.data["url"],
                    "description": f"Duplicate Host Header Tolerated",
                },
                "FINDING",
                event,
            )

        # host header overrides
        technique_description = "host override headers"
        self.verbose(f"Performing {technique_description} case")
        subdomain_tag = self.rand_string(4, digits=False)
        self.subdomain_tags[subdomain_tag] = (event, technique_description)

        override_headers_list = [
            "X-Host",
            "X-Forwarded-Server",
            "X-Forwarded-Host",
            "X-Original-Host",
            "X-Forwarded-For",
            "X-Host",
            "X-HTTP-Host-Override",
            "Forwarded",
        ]
        override_headers = {}
        for oh in override_headers_list:
            override_headers[oh] = f"{subdomain_tag}.{self.domain}"

        output = await self.helpers.curl(
            url=event.data["url"],
            headers=override_headers,
            cookies=added_cookies,
        )
        if self.domain in output:
            domain_reflections.append(technique_description)

        # emit all the domain reflections we found
        for dr in domain_reflections:
            await self.emit_event(
                {
                    "host": str(event.host),
                    "url": event.data["url"],
                    "description": f"Possible Host header injection. Injection technique: {dr}",
                },
                "FINDING",
                event,
            )



#!/usr/bin/env python3

import os
import re
import yaml
from pathlib import Path

from bbot.modules import module_loader
from bbot.core.configurator.args import parser, scan_examples

os.environ["BBOT_TABLE_FORMAT"] = "github"


# Make a regex pattern which will match any group of non-space characters that include a blacklisted character
blacklist_chars = ["<", ">"]
blacklist_re = re.compile(r"\|([^|]*[" + re.escape("".join(blacklist_chars)) + r"][^|]*)\|")

bbot_code_dir = Path(__file__).parent.parent.parent


def enclose_tags(text):
    # Use re.sub() to replace matched words with the same words enclosed in backticks
    result = blacklist_re.sub(r"|`\1`|", text)
    return result


def find_replace_markdown(content, keyword, replace):
    begin_re = re.compile(r"<!--\s*" + keyword + r"\s*-->", re.I)
    end_re = re.compile(r"<!--\s*END\s+" + keyword + r"\s*-->", re.I)

    begin_match = begin_re.search(content)
    end_match = end_re.search(content)

    new_content = str(content)
    if begin_match and end_match:
        start_index = begin_match.span()[-1] + 1
        end_index = end_match.span()[0] - 1
        new_content = new_content[:start_index] + enclose_tags(replace) + new_content[end_index:]
    return new_content


def find_replace_file(file, keyword, replace):
    with open(file) as f:
        content = f.read()
        new_content = find_replace_markdown(content, keyword, replace)
    if new_content != content:
        if not "BBOT_TESTING" in os.environ:
            with open(file, "w") as f:
                f.write(new_content)


def update_docs():
    md_files = [p for p in bbot_code_dir.glob("**/*.md") if p.is_file()]

    def update_md_files(keyword, s):
        for file in md_files:
            find_replace_file(file, keyword, s)

    def update_individual_module_options():
        regex = re.compile("BBOT MODULE OPTIONS ([A-Z_]+)")
        for file in md_files:
            with open(file) as f:
                content = f.read()
            for match in regex.finditer(content):
                module_name = match.groups()[0].lower()
                bbot_module_options_table = module_loader.modules_options_table(modules=[module_name])
                find_replace_file(file, f"BBOT MODULE OPTIONS {module_name.upper()}", bbot_module_options_table)

    # Example commands
    bbot_example_commands = []
    for title, description, command in scan_examples:
        example = ""
        example += f"**{title}:**\n\n"
        # example += f"{description}\n"
        example += f"```bash\n# {description}\n{command}\n```"
        bbot_example_commands.append(example)
    bbot_example_commands = "\n\n".join(bbot_example_commands)
    assert len(bbot_example_commands.splitlines()) > 10
    update_md_files("BBOT EXAMPLE COMMANDS", bbot_example_commands)

    # Help output
    bbot_help_output = parser.format_help().replace("docs.py", "bbot")
    bbot_help_output = f"```text\n{bbot_help_output}\n```"
    assert len(bbot_help_output.splitlines()) > 50
    update_md_files("BBOT HELP OUTPUT", bbot_help_output)

    # BBOT events
    bbot_event_table = module_loader.events_table()
    assert len(bbot_event_table.splitlines()) > 10
    update_md_files("BBOT EVENTS", bbot_event_table)

    # BBOT modules
    bbot_module_table = module_loader.modules_table()
    assert len(bbot_module_table.splitlines()) > 50
    update_md_files("BBOT MODULES", bbot_module_table)

    # BBOT module options
    bbot_module_options_table = module_loader.modules_options_table()
    assert len(bbot_module_options_table.splitlines()) > 100
    update_md_files("BBOT MODULE OPTIONS", bbot_module_options_table)
    update_individual_module_options()

    # BBOT module flags
    bbot_module_flags_table = module_loader.flags_table()
    assert len(bbot_module_flags_table.splitlines()) > 10
    update_md_files("BBOT MODULE FLAGS", bbot_module_flags_table)

    # Default config
    default_config_file = bbot_code_dir / "bbot" / "defaults.yml"
    with open(default_config_file) as f:
        default_config_yml = f.read()
    default_config_yml = f'```yaml title="defaults.yml"\n{default_config_yml}\n```'
    assert len(default_config_yml.splitlines()) > 20
    update_md_files("BBOT DEFAULT CONFIG", default_config_yml)

    # Table of Contents
    base_url = "https://www.blacklanternsecurity.com/bbot"

    def format_section(section_title, section_path):
        path = section_path.split("index.md")[0]
        path = path.split(".md")[0]
        return f"- [{section_title}]({base_url}/{path})\n"

    bbot_docs_toc = ""

    def update_toc(section, level=0):
        nonlocal bbot_docs_toc
        indent = " " * 4 * level
        if isinstance(section, dict):
            for section_title, subsections in section.items():
                if isinstance(subsections, str):
                    bbot_docs_toc += f"{indent}{format_section(section_title, subsections)}"
                else:
                    bbot_docs_toc += f"{indent}- **{section_title}**\n"
                    for subsection in subsections:
                        update_toc(subsection, level=level + 1)

    mkdocs_yml_file = bbot_code_dir / "mkdocs.yml"
    yaml.SafeLoader.add_constructor(
        "tag:yaml.org,2002:python/name:pymdownx.superfences.fence_code_format", lambda x, y: {}
    )

    with open(mkdocs_yml_file, "r") as f:
        mkdocs_yaml = yaml.safe_load(f)
        nav = mkdocs_yaml["nav"]
        for section in nav:
            update_toc(section)
    bbot_docs_toc = bbot_docs_toc.strip()
    # assert len(bbot_docs_toc.splitlines()) == 2
    update_md_files("BBOT DOCS TOC", bbot_docs_toc)


update_docs()



from bbot.db.neo4j import Neo4j
from bbot.modules.output.base import BaseOutputModule


class neo4j(BaseOutputModule):
    """
    docker run -p 7687:7687 -p 7474:7474 -v "$(pwd)/data/:/data/" -e NEO4J_AUTH=neo4j/bbotislife neo4j
    """

    watched_events = ["*"]
    meta = {"description": "Output to Neo4j"}
    options = {"uri": "bolt://localhost:7687", "username": "neo4j", "password": "bbotislife"}
    options_desc = {
        "uri": "Neo4j server + port",
        "username": "Neo4j username",
        "password": "Neo4j password",
    }
    deps_pip = ["git+https://github.com/blacklanternsecurity/py2neo"]
    _batch_size = 50
    _preserve_graph = True

    async def setup(self):
        try:
            self.neo4j = await self.scan.run_in_executor(
                Neo4j,
                uri=self.config.get("uri", self.options["uri"]),
                username=self.config.get("username", self.options["username"]),
                password=self.config.get("password", self.options["password"]),
            )
            await self.scan.run_in_executor(self.neo4j.insert_event, self.scan.root_event)
        except Exception as e:
            self.warning(f"Error setting up Neo4j: {e}")
            return False
        return True

    async def handle_event(self, event):
        await self.scan.run_in_executor(self.neo4j.insert_event, event)

    async def handle_batch(self, *events):
        await self.scan.run_in_executor(self.neo4j.insert_events, events)
